define([
  'dojo/query',
  'dojo/dom',
  'dojo/dom-class',
  'dojo/dom-style',
  'dojo/dom-attr',
  'dojo/topic',
  'dojo/has',
  'dojo/on',
  'dojo/_base/window',
  'dojo/_base/array',
  'esri/geometry/Extent',
  'esri/SpatialReference',
  'esri/layers/GraphicsLayer',
  'dojo/_base/lang',
  './config',
  './mapping/MapControls',
  './mapping/ContentGroup',
  './layout/NavBar',
  './layout/IFrameModal',
  './layout/IFrameModalBrowserAdvice',
  './layout/MapLegend',
  './layout/PortalGlossary',
  './utils/Bookmarker',
  './search/SearchEngine',
  './search/SearchResults',
  './mapping/Utils',
  'dojo-bootstrap/Modal',
  'dojo-bootstrap/Collapse',
  './layout/CollapseNext',
  'dojo/domReady!'],
function (
  query, dom, domClass, domStyle, domAttr, topic, has, on, win, array, Extent, SpatialReference, GraphicsLayer, lang, config, MapControls, ContentGroup, NavBar, IFrameModal, IFrameModalBrowserAdvice, MapLegend, PortalGlossary, Bookmarker, SearchEngine, SearchResults, mapUtils
) {
    'use strict';
    var app = {};
    
    // display browser warning for IE 9 and lower
    if (has('ie') <= 9) {
        app.iframeModalBrowserAdvice = new IFrameModalBrowserAdvice({ iframeId: 'iframeModalAdviceDiv' }, 'iframeModalAdviceContent');
        app.iframeModalBrowserAdvice.startup();
        query('[id="iframeModalAdviceDiv"]').modal('show');
    }

    // start map
    app.mapControls = new MapControls(config.mapControls, 'mapControls');
    app.mapControls.startup();

    app.sideBar = query('#sidebar');

    var searchLayer = new GraphicsLayer();
    app.mapControls.map.addLayer(searchLayer, 0);

    //subscribe to ExecuteOnMap
    topic.subscribe('map/execute', function (delegate) {
        delegate(app.mapControls.map);
    });

    topic.subscribe('map/showSearchLayer', function (graphic) {
        searchLayer.clear();
        searchLayer.add(graphic);
        searchLayer.show();
        topic.publish('map/reorderlayers');
    });

    topic.subscribe('map/hideSearchLayer', function () {
        searchLayer.hide();
    });

    topic.subscribe('map/reorderlayers', function () {
        mapUtils.reorderLayers(app.mapControls.map);
    });

    topic.subscribe('map/layers/show-by-layer-name', function(layerName) {
        var nodes = query("[data-layer-name*='" + layerName + "']:not(.active)", app.sidebar);
        nodes.forEach(function(node) {
            if (!domClass.contains(node, 'active')) {
                node.click();
            }
        });

        var parentCollapsePanels = query(nodes[0]).parents('.collapse');
        array.forEach(parentCollapsePanels.reverse(), function(parentCollapsePanel) {
            query("[id='" + parentCollapsePanel.id + "']", app.sidebar).collapse('show');
        });
    });

    topic.subscribe('map/layers/hide-by-layer-name', function(layerName) {
        var nodes = query("[data-layer-name*='" + layerName + "']", app.sidebar);
        nodes.forEach(function(node) {
            if (domClass.contains(node, 'active')) {
                node.click();
            }
        });
    });
    

    // start nav
    app.navBar = new NavBar({}, 'navBar');
    app.navBar.set('titleLink', siteConfig.logoUrl);
    app.navBar.set('homeLink', siteConfig.rootUrl + 'Map/');
    app.navBar.set('libraryLink', siteConfig.libraryUrl);
    app.navBar.set('logoImageSrc', siteConfig.rootUrl + 'Images/AB-Gov-Reverse.png');
    app.navBar.set('logoImageMobileSrc', siteConfig.rootUrl + 'Images/AB-Gov-Reverse.png');
    app.navBar.set('siteLinks', config.siteLinks);
    app.navBar.set('aemerisLink', siteConfig.aemerisMonitoringDataUrl);
    app.navBar.startup();

    array.forEach(config.contentGroups.group, function (contentGroupConfig) {
        // add Content group
        app.contentGroup = new ContentGroup(
            contentGroupConfig,
            app.mapControls,
            app.mapControls.map,
            config.mapControls.options.extent,
            app.mapControls.contentGroupNodeId);
    });

    // reorder the layers
    topic.publish('map/reorderlayers');

    //Search
    app.searchEngine = SearchEngine;
    app.searchResults = new SearchResults({}, 'searchResultsModalContent');
    app.searchResults.startup();

    // create map legend
    app.MapLegend = new MapLegend({
        theMap: app.mapControls.map,
        contentGroups: config.contentGroups,
        legendNodeId: app.navBar.legendWidgetNode,
        operationLayers: config.mapControls.operationalLayers
    });
    app.MapLegend.startup();

    topic.subscribe("UpdateState/LayerChange", function () {
        app.updateState();
    });

    // iframe for charts and other pages
    app.iframeModal = new IFrameModal({ iframeId: 'modalFrame' }, 'iframeModalContent');
    app.iframeModal.startup();
    
    // modal for glossay
    app.modalPortalGlossary = new PortalGlossary({}, 'glossaryModalContent');
    app.modalPortalGlossary.startup();

    // responsive sidebar
    app.sidebar = dom.byId('sidebar');

    // check if map is within max extent when extent changes.  If not, roll back to max extent
    app.mapControls.map.on("extent-change", function (event) {
       
        if ((event.extent.xmin < config.mapControls.options.maxExtent.xmin) ||
           (event.extent.ymin < config.mapControls.options.maxExtent.ymin) ||
           (event.extent.xmax > config.mapControls.options.maxExtent.xmax) ||
           (event.extent.ymax > config.mapControls.options.maxExtent.ymax)) {
            app.mapControls.map.setExtent(config.mapControls.options.extent);
        }     
    });

    // app topics
    // set app title and about modal based on web map
    topic.subscribe('webmap/load', function (args) {
        var item, content;
        if (!args.itemInfo || !args.itemInfo.item) {
            return;
        }
        item = args.itemInfo.item;
        app.navBar.set('title', item.title);
        content = 'This sample is a modified version of the <a href="https://github.com/Esri/dojo-bootstrap-map-js">dojo-bootstrap-map-js</a> <a href="http://esri.github.io/dojo-bootstrap-map-js/">demo page</a>.<br/>';
        content += item.snippet || item.description;
    });

    // toggle the sidebar
    topic.subscribe('sidebar/toggle', function () {
        if (!app.sidebar) {
            return;
        }
        // make sure sidebar is same height as the map
        domStyle.set(app.sidebar, 'height', app.mapControls.getMapHeight() + 'px');
        domClass.toggle(window.document.body, 'sidebar-open');
    });

    // show the about modal
    topic.subscribe('about/show', function () {
        query('.about-modal').modal('show');
    });

    topic.subscribe('mapcontrols/loaded', function () {
        app.bookmarker = new Bookmarker({
            onStateChange: lang.hitch(app, app.onWindowStateChange),
            restoreState: lang.hitch(app, app.restoreState),
            encode: false,
            getState: lang.hitch(app, app.getState)
        });
        topic.subscribe('state/change', lang.hitch(app, function () {
            this.bookmarker.saveState();
        }));
        app.mapControls.map.on('pan-end, zoom-end', app.updateState);

        app.mapControls.basemapGallery.on('selection-change', function () {
            app.updateState();
            topic.publish('map/reorderlayers');
        });

        //app.mapControls.map.on('basemap-change', function () {
        //    console.log("map basemap-change");
        //    app.updateState();
        //});
        //app.mapControls.map.on('pan-end', function() {
        //    console.log("map pan-end");
        //    app.updateState();
        //});
        //app.mapControls.map.on('zoom-end', function() {
        //    console.log("map zoom-end");
        //    app.updateState();
        //});
        app.mapControls.map.on('layer-add', function (addEvent) {
            //save state when layer visibility changes
            addEvent.layer.on('visibility-change', function () {
                app.updateState();
                topic.publish('map/reorderlayers');
            });
        });
        //var handle = app.mapControls.map.on('update-end', function () {           

        //    //on initial page load get state from querystring
        //    var bookmarkedState = app.bookmarker.getStateFromQueryString();
        //    if (bookmarkedState) {
        //        app.bookmarker.restoreState(bookmarkedState);
        //    } else {
        //        app.bookmarker.cleanLoad = false;
        //    }

        //    handle.remove();
        //    //console.log("update-end completed.... map should be done updating...");
        //});
        var bookmarkedState = app.bookmarker.getStateFromQueryString();
        if (bookmarkedState) {
            app.bookmarker.restoreState(bookmarkedState);
        } else {
            app.bookmarker.cleanLoad = false;
        }
    });

    topic.subscribe('map/layerControlClickHandler', function (event) {
        mapUtils.layerControlClickHandler(event, app.mapControls.map);
    });

    app.getState = function () {
        //This function is where we save items to the state
        var extent = this.mapControls.map.extent,
            basemap = this.mapControls.basemapGallery.getSelected(),
            visibleLayers = this.mapControls.map.getLayersVisibleAtScale(this.mapControls.map.getScale()).filter(function(layer) { return layer.visible; }),
            visibleLayersWithNoSubLayers = [],
            visibleLayersWithSubLayers = [];

        visibleLayers.forEach(function (layer) {
            if (layer.visibleLayers && layer.visibleLayers.length) {
                visibleLayersWithSubLayers.push({ layerId: layer.id, subLayerIds: layer.visibleLayers });
            } else {
                visibleLayersWithNoSubLayers.push(layer.id);
            }
        });

        var stateObj = {
            extent: {
                xmin: extent.xmin,
                ymin: extent.ymin,
                xmax: extent.xmax,
                ymax: extent.ymax,
                spatialReference: { wkid: extent.spatialReference.wkid }
            },
            basemap: basemap !== undefined && basemap !== null ? basemap.id : this.mapControls.options.basemap,
            visibleLayersWithNoSubLayers: visibleLayersWithNoSubLayers
        };

        if (visibleLayersWithSubLayers.length) {
            stateObj.visibleLayersWithSubLayers = visibleLayersWithSubLayers;
        }

        return stateObj;
    };

    app.restoreState = function (state) {
        //This function is where we restore items from the state
        var currentLayers, layersToHide, nodesTobeEnabled = [],
            hasExtent = Number.isNaN ? !Number.isNaN(app.mapControls.map.extent.xmin) : !isNaN(app.mapControls.map.extent.xmin); 
        app.bookmarker.justRestoredState = true;

        if (hasExtent && ((app.mapControls.map.extent.xmin !== state.extent.xmin && app.mapControls.map.extent.ymin !== state.extent.ymin) ||
            (app.mapControls.map.extent.xmax !== state.extent.xmax && app.mapControls.map.extent.ymax !== state.extent.ymax))) {
            app.mapControls.map.setExtent(new Extent(state.extent.xmin, state.extent.ymin, state.extent.xmax, state.extent.ymax, new SpatialReference({ wkid: state.extent.spatialReference.wkid })));
        }
        if (state.visibleLayersWithNoSubLayers) {
            currentLayers = this.mapControls.map.getLayersVisibleAtScale(this.mapControls.map.getScale());
            layersToHide = currentLayers.filter(function (layer) {
                //if layer is visible and not in the new state visibleLayers list then return true to hide it.
                return layer.visible && state.visibleLayersWithNoSubLayers.indexOf(layer.id) === -1;
            });
            layersToHide.forEach(function (layer) {
                //need to set justRestoredState to true before hiding or showing a layer while restoring to prevent saving the state.
                //this is because justRestoredState gets set to false at the end of each save state call.
                app.bookmarker.justRestoredState = true;
                layer.hide();
            });

            state.visibleLayersWithNoSubLayers.forEach(function (layerId) {
                // expand the collapse and check the corresponding list item
                query("[data-layer-id='" + layerId + "']").forEach(function (node) {
                    nodesTobeEnabled.push(node);
                });
            });
        }
        
        if (state.visibleLayersWithSubLayers) {
            state.visibleLayersWithSubLayers.forEach(function (e) {
                query("[data-layer-id*='" + e.layerId + "']").forEach(lang.hitch(e, function (node) {
                    var subLayerIdsInState = e.subLayerIds.join(), // convert it to a string for comparison
                        subLayerIdsInNode = domAttr.get(node, 'data-visible-layers');

                    if (subLayerIdsInState.indexOf(subLayerIdsInNode) >= 0) { // node sublayerIds exist in state layer Ids
                        nodesTobeEnabled.push(node);
                    }                  
                }));
            });
        }
      
        // enable all control nodes
        setTimeout(function () {
            nodesTobeEnabled.forEach(function (node) {

                var collapsePanel = node.parentNode.parentNode;

                if (node.nodeName === 'OPTION') {
                    // set historical year
                    node.parentNode.value = node.value;
                    // trigger onchange event
                    var event = document.createEvent('Event');
                    node.parentNode.onchange(event);
                    collapsePanel = app.findCollapsibleAncestor(node);
                }
                else {
                    node.click();
                }

                if (node.parentNode.parentNode) {
                    if (domClass.contains(collapsePanel, "collapse")) {
                        // expand the collapse control that contains the checked list item
                        query("[id='" + collapsePanel.id + "']").collapse('show');
                    }

                    // check whether the panel has a parent collapse panel and expand it
                    if (collapsePanel.attributes.parentNodeId !== undefined) {
                        app.expandParentNode(collapsePanel.attributes.parentNodeId.value);
                    }
                }
                    
            });
            // reorder the layers
            topic.publish('map/reorderlayers');
        }, 500);
        if (state.basemap) {
            app.mapControls.basemapGallery.select(state.basemap);
        }

        app.bookmarker.cleanLoad = false;
    };

    app.findCollapsibleAncestor = function (node) {
        if (!domClass.contains(node.parentNode, "collapse")) {
            return app.findCollapsibleAncestor(node.parentNode);
        }
        //Return the node if it contains collapse or undefined if it doesn't exist
        return node.parentNode;
    };

    // expand parent node
    app.expandParentNode = function (parentNodeID) {
        var parentNode = dom.byId(parentNodeID);
        if (domClass.contains(parentNode, "collapse")) {
            // expand the collapse control that contains the checked list item
            query("[id='" + parentNode.id + "']").collapse('show');
        }

        if(parentNode.attributes.parentNodeId !== undefined)
        {
            app.expandParentNode(parentNode.attributes.parentNodeId.value);
        }
    };

    app.updateState = function () {
        topic.publish('state/change');
    };

    app.onWindowStateChange = function () {
        if (!app.bookmarker.justSavedState && !app.bookmarker.cleanLoad) {
            var state = History.getState();
            app.restoreState(state.data);//state object from History contains additional stuff and we only need the data component
            app.bookmarker.justRestoredState = true;
        } else {
            app.bookmarker.justSavedState = false;
            app.bookmarker.cleanLoad = false;
        }
    };
    
    // listen for click events on elements which has "layerControl" class name
    on(win.doc, ".layerControl:click", function (event) {
        topic.publish('map/layerControlClickHandler', event);
    });

   
    return app;
});
