﻿define([
    'dojo/topic',
    'dojo/_base/array',
    'dojo/_base/declare',
    'dojo/_base/lang',
    'dojo/string',
    'dojo/promise/all',
    'dojo/Deferred',
    'dojo/request/xhr',
    './_searchBase',
    'app/config',
    '../SearchableConfig',
    'dojo/i18n!app/layout/nls/strings',
    'esri/tasks/query',
    'esri/tasks/QueryTask',
    'esri/SpatialReference',
    'esri/graphic',
    'esri/symbols/SimpleFillSymbol', 
    'esri/symbols/SimpleLineSymbol', 
    'esri/Color',
    'esri/geometry/webMercatorUtils',
    'esri/geometry/Point',
    'esri/symbols/PictureMarkerSymbol'
], function (topic, array, declare, lang, string, all, Deferred, xhr, searchBase, config, SearchableConfig, strings,
    Query, QueryTask, SpatialReference, Graphic, SimpleFillSymbol, SimpleLineSymbol, Color, webMercatorUtils, Point, PictureMarkerSymbol) {
    'use strict';
    return declare(searchBase, {

        constructor: function () {
            this.resultsType = strings.searchMapLayerOSIPProjectResultsTitle;
            this.oilSandsProjectBoundariesLayerSearchUrl = JSON.search(new SearchableConfig(), "//contentGroups/group/operationalLayers/subGroups/operationalLayers/subGroups/operationalLayers/options[id='projectBoundaries']/../url")[0];
            if (this.oilSandsProjectBoundariesLayerSearchUrl) {
                this.searchQueryTask = new QueryTask(this.oilSandsProjectBoundariesLayerSearchUrl);
            } else {
                throw 'Could not find oil sands project boundaries layer search url in config.';
            }
        },

        getWhere: function (searchText) {
            return string.substitute('Project_Name like \'%${searchText}%\' OR Operator_Name like \'%${searchText}%\'', { searchText: searchText });
        },

        queryMapServer: function (searchText) {

            var deferred = new Deferred();

            var layerQuery = new Query();
            layerQuery.returnGeometry = true;
            layerQuery.outSpatialReference = new SpatialReference(102100);
            layerQuery.outFields = this.getOutFields;
            layerQuery.where = this.getWhere(searchText);

            this.searchQueryTask.execute(layerQuery, function (response) {
                deferred.resolve(response.features);
            });

            return deferred;
        },

        queryOsipApi: function (searchText) {

            var deferred = new Deferred();

            var osipQueryUrl = string.substitute(config.oilSandsApiUrl + '/Projects?$expand=Operators&$filter=Operators/any(o: substringof(\'${searchText}\', tolower(o/Name)) eq true) or substringof(\'${searchText}\', tolower(Name)) eq true and IsShownOnMap eq true', { searchText: searchText.toLowerCase() });

            xhr(osipQueryUrl, {
                sync: false,
                handleAs: 'json',
                headers: { 'X-Requested-With': null }
            }).then(function (apiResponse) {
                deferred.resolve(apiResponse.value);
            });

            return deferred;
        },

        getOutFields: ['*'],

        searchAsync: function (searchText) {
            
            var resultsDeferred = new Deferred();

            var deferredMapLayerResults = this.queryMapServer(searchText);
            var deferredOsipApiResults = this.queryOsipApi(searchText);

            var self = this;
            
            all({ MapLayerResults: deferredMapLayerResults, OsipApiProjects: deferredOsipApiResults }).then(lang.hitch(self, function (results) {

                var combinedResults = {};

                array.forEach(results.MapLayerResults, function (mapLayerResult) {
                    combinedResults[String(mapLayerResult.attributes.OSP_NO)] = { MapLayerResult: mapLayerResult };
                });

                array.forEach(results.OsipApiProjects, function (osipApiProject) {
                    if (combinedResults[String(osipApiProject.ProjectIdentifier)]) {
                        combinedResults[String(osipApiProject.ProjectIdentifier)].OsipApiProject = osipApiProject;
                    } else {
                        combinedResults[String(osipApiProject.ProjectIdentifier)] = { OsipApiProject: osipApiProject };
                    }
                });

                var resultsArray = [];
                Object.keys(combinedResults).forEach(function(projectId) {
                    resultsArray.push(combinedResults[projectId]);
                });

                resultsDeferred.resolve(this.getResultModels(resultsArray, searchText));
            }));

            return resultsDeferred;
        },

        searchModuleIdentifier: 'MapLayerOSIPProjectPoint', 

        getResultModel: function (resultObject, searchText) {
            var lowerCaseSearchText = searchText.toLowerCase();
            var matchedField = strings.searchFieldUnknown;
            var matchedValue;
            var title;

            if (resultObject.OsipApiProject) {

                if (resultObject.OsipApiProject.Name.toLowerCase().indexOf(lowerCaseSearchText) > -1) {
                    matchedValue = resultObject.OsipApiProject.Name;
                    matchedField = strings.searchMapLayerOSIPProjectResultsProjectNameField;
                } else if (resultObject.OsipApiProject.Operators) {
                    array.forEach(resultObject.OsipApiProject.Operators, function (operator) {
                        if (operator.Name.toLowerCase().indexOf(lowerCaseSearchText) > -1) {
                            matchedValue = operator.Name;
                            matchedField = strings.searchMapLayerOSIPProjectResultsOperatorNameField;
                        }
                    });
                }
                title = resultObject.OsipApiProject.Name + ' - ' + resultObject.OsipApiProject.Operators[0].Name ;
            } else if (resultObject.MapLayerResult) {

                if (resultObject.MapLayerResult.attributes.Project_Name.toLowerCase().indexOf(lowerCaseSearchText) > -1) {
                    matchedValue = resultObject.MapLayerResult.attributes.Project_Name;
                    matchedField = strings.searchMapLayerOSIPProjectResultsProjectNameField;
                } else if (resultObject.MapLayerResult.attributes.Operator_Name.toLowerCase().indexOf(lowerCaseSearchText) > -1) {
                    matchedValue = resultObject.MapLayerResult.attributes.Operator_Name;
                    matchedField = strings.searchMapLayerOSIPProjectResultsOperatorNameField;
                }
                title = resultObject.MapLayerResult.attributes.Project_Name;
            } else {
                throw 'Unknown Result Type';
            }
            
            var navigateAction;

            if (resultObject.MapLayerResult) {
                navigateAction = function() {
                    topic.publish('map/layers/show-by-layer-name', 'projectOilSands');
                    topic.publish('map/execute', function(map) {
                        var extent = resultObject.MapLayerResult.geometry.getExtent();
                        map.centerAt(extent.getCenter());
                        var stateExtent = extent.expand(10.0);
                        map.setExtent(stateExtent);
                        var sfs = new SimpleFillSymbol(SimpleFillSymbol.STYLE_SOLID,
                            new SimpleLineSymbol(SimpleLineSymbol.STYLE_DASHDOT,
                                new Color([255, 0, 0]), 2), new Color([255, 0, 0, 0.25])
                        );
                        topic.publish('map/showSearchLayer', new Graphic(resultObject.MapLayerResult.geometry, sfs));
                    });
                };
            } else if (resultObject.OsipApiProject) {
                navigateAction = function() {
                    topic.publish('map/layers/show-by-layer-name', 'projectOilSands');
                    topic.publish('map/execute', function(map) {
                        var point = webMercatorUtils.geographicToWebMercator(
                            new Point(resultObject.OsipApiProject.Longitude, resultObject.OsipApiProject.Latitude));

                        map.centerAndZoom(point, 10);
                        var symbol = new PictureMarkerSymbol(require.toUrl("app/content/images/projects/Project_Search_Highlight.png"), 34, 32);
                        topic.publish('map/showSearchLayer', new Graphic(point, symbol));
                    });
                };
            } else {
                navigateAction = function () { throw 'Navigation for OSIP Project Result Unknown'; };
            }

            return {
                resultType: this.resultsType,
                title: title,
                navigateAction: navigateAction,
                field: matchedField,
                matchedValue: matchedValue,
                glyphIconClass: 'glyphicon-map-marker'
            };            
        }
    });
});