﻿define([
    'intern!object',
    'intern/chai!assert',
    'dojo/_base/array',
    'dojo/_base/Color',
    'app/mapping/Utils',
    'esri/geometry/Extent',
    'esri/graphic',
    'esri/layers/GraphicsLayer',
    'esri/symbols/SimpleLineSymbol',
    'esri/symbols/SimpleFillSymbol',
    'esri/layers/FeatureLayer',
    'esri/geometry/Polygon'
], function (registerSuite, assert, array, Color, mapUtils, Extent, Graphic, GraphicsLayer, SimpleLineSymbol, SimpleFillSymbol, FeatureLayer, Polygon) {
    'use strict';
    registerSuite({
        name: 'Map Utilities module tests',
        'Test Create Graphics Layer': function () {
            var layerId = 1,
                layer = mapUtils.createGraphicsLayer(layerId, true);

            assert.isTrue(layer instanceof GraphicsLayer, "Create Graphics Layer doesn't return expected object");
            assert.strictEqual(layer.id, layerId, "Create Graphics Layer doesn't return expected layer ID");
            assert.isTrue(layer.visible, "Create Graphics Layer doesn't return expected layer visible");

            layerId = 100;
            layer = mapUtils.createGraphicsLayer(layerId, false);

            assert.isTrue(layer instanceof GraphicsLayer, "Create Graphics Layer doesn't return expected object");
            assert.strictEqual(layer.id, layerId, "Create Graphics Layer doesn't return expected layer ID");
            assert.isFalse(layer.visible, "Create Graphics Layer doesn't return expected layer visible");
        },
        'Test Query Feature Layer': function () {
            // ReSharper disable once JoinDeclarationAndInitializerJs
            var def = this.async(7000),
                layerUrl = 'https://maps.alberta.ca/genesis_tokenauth/rest/services/OSIP-Layers/20161021/MapServer/6',
                outFields = ['OBJECTID', 'OSP_NO'],
                hasValidFeature = false,
                iniExtent = new Extent({
                    xmin: -14265055.060527408,
                    ymin: 6259094.656300137,
                    xmax: -11283399.461180044,
                    ymax: 8409115.387905004,
                    spatialReference: {
                        wkid: 102100
                    }
                }),
                successCallback = function (data) {
                    if (data.features.length === 0) {
                        assert.notStrictEqual(data.features.length, 0, "Query Feature Layer doesn't any data" );
                        def.reject();
                    } else {
                        hasValidFeature = array.some(data.features, function (feature) {
                            if (feature.attributes.OSP_NO > 0) {
                                return true;
                            }
                        });
                        assert.isTrue(hasValidFeature, "Query Feature Layer doesn't return valid features.");
                        def.resolve(data);
                    }                
                },
                errorCallback = function (error) {
                    assert.fail("Query Feature Layer failed: " + error);
                    def.reject();
                };

            mapUtils.queryFeatureLayer(successCallback, errorCallback, layerUrl, outFields, iniExtent);
        },
        'Test Get SimpleFillSymbol': function () {
            var expectedFillTransparency = 1,
                expectedFillColor = new Color([100, 100, 100, expectedFillTransparency]),
                expectedBorderTransparency = 0.25,
                expectedBorderColour = new Color([150, 150, 150, expectedBorderTransparency]),
                expectedBorderWidth = 2,
                result = mapUtils.getSimpleFillSymbol(expectedFillColor, expectedFillTransparency, expectedBorderColour, expectedBorderTransparency, expectedBorderWidth);

            assert.isTrue(result instanceof SimpleFillSymbol,
                "Get SimpleFillSymbol doesn't return expected symbol");
            assert.strictEqual(JSON.stringify(result.color), JSON.stringify(expectedFillColor),
                "Get SimpleFillSymbol doesn't return expected fill color");
            assert.strictEqual(result.color.a, expectedFillTransparency,
                "Get SimpleFillSymbol doesn't return expected fill transparency");
            assert.strictEqual(JSON.stringify(result.outline.color), JSON.stringify(expectedBorderColour),
                "Get SimpleFillSymbol doesn't return expected border color");
            assert.strictEqual(result.outline.color.a, expectedBorderTransparency,
                "Get SimpleFillSymbol doesn't return expected border transparency");
            assert.strictEqual(result.outline.width, expectedBorderWidth,
               "Get SimpleFillSymbol doesn't return expected border width");

            // test with different values
            expectedFillTransparency = 0.35;
            expectedFillColor = new Color([125, 80, 210, expectedFillTransparency]);
            expectedBorderTransparency = 0.75;
            expectedBorderColour = new Color([200, 30, 140, expectedBorderTransparency]);
            expectedBorderWidth = 3;
            result = mapUtils.getSimpleFillSymbol(expectedFillColor, expectedFillTransparency, expectedBorderColour, expectedBorderTransparency, expectedBorderWidth);

            assert.isTrue(result instanceof SimpleFillSymbol,
                "Get SimpleFillSymbol doesn't return expected symbol");
            assert.strictEqual(JSON.stringify(result.color), JSON.stringify(expectedFillColor),
                "Get SimpleFillSymbol doesn't return expected fill color");
            assert.strictEqual(result.color.a, expectedFillTransparency,
                "Get SimpleFillSymbol doesn't return expected fill transparency");
            assert.strictEqual(JSON.stringify(result.outline.color), JSON.stringify(expectedBorderColour),
                "Get SimpleFillSymbol doesn't return expected border color");
            assert.strictEqual(result.outline.color.a, expectedBorderTransparency,
                "Get SimpleFillSymbol doesn't return expected border transparency");
            assert.strictEqual(result.outline.width, expectedBorderWidth,
               "Get SimpleFillSymbol doesn't return expected border width");

        },
        'Test Add GraphicFeature To Layer': function () {
            var layerDefinition = {
                geometryType: "esriGeometryPolygon",
                fields: [
                    {
                        name: "id",
                        type: "esriFieldTypeString"
                    },
                    {
                        name: "stringField",
                        type: "esriFieldTypeString"
                    },
                    {
                        name: "doubleField",
                        type: "esriFieldTypeDouble"
                    }
                ]
            },
                featureCollection = {
                    layerDefinition: layerDefinition,
                    featureSet: null
                },
                layer = new FeatureLayer(featureCollection, { id: "testId", visible: false, mode: FeatureLayer.MODE_SNAPSHOT }),
                polygonJson = {
                    rings: [[[-122.63, 45.52], [-122.57, 45.53], [-122.52, 45.50], [-122.49, 45.48],
                [-122.64, 45.49], [-122.63, 45.52], [-122.63, 45.52]]], spatialReference: { wkid: 4326 }
                },
                polygon = new Polygon(polygonJson),
                graphicData = {
                    geometry: polygon,
                    attributes: { id: null, stringField: "Pond 1", doubleField: 48.54 },
                    symbol: { color: [0, 0, 0, 255], width: 1, type: "esriSLS", style: "esriSLSSolid" },
                    infoTemplate: { title: "test title", content: "${*}" }
                },
                graphic = new Graphic(graphicData),
                idHelper = 'BBD8220F-886E-A8C2-8F2A-CDF17F081BB9',
                symbol = new SimpleFillSymbol(SimpleFillSymbol.STYLE_SOLID,
                    new SimpleLineSymbol(
                        SimpleLineSymbol.STYLE_SOLID,
                        new Color([150, 150, 150, 0.75]),
                        3),
                    new Color([100, 100, 100, 0.25]));

            mapUtils.addGraphicFeatureToLayer(layer, graphic, symbol, idHelper);

            assert.strictEqual(layer.graphics.length, 1, "Add GraphicFeature To Layer doesn't save the graphic object.");
            assert.isTrue(layer.graphics[0] instanceof Graphic, "Add GraphicFeature To Layer doesn't save expected object.");
            assert.deepEqual(layer.graphics[0].geometry.rings, polygon.rings, "Add GraphicFeature To Layer doesn't save expected geometry paths.");
            assert.deepEqual(layer.graphics[0].geometry.spatialReference.wkid, polygon.spatialReference.wkid,
                "Add GraphicFeature To Layer doesn't save expected geometry spatialReference.");
            assert.deepEqual(layer.graphics[0].symbol, symbol,
                "Add GraphicFeature To Layer doesn't save expected symbol.");
            assert.strictEqual(layer.graphics[0].attributes.id, idHelper, "Add GraphicFeature To Layer doesn't save expected idHelper.");
            assert.strictEqual(layer.graphics[0].attributes.stringField, graphicData.attributes.stringField, "Add GraphicFeature To Layer doesn't save expected POND_NAME.");
            assert.strictEqual(layer.graphics[0].attributes.doubleField, graphicData.attributes.doubleField, "Add GraphicFeature To Layer doesn't save expected HECTARES.");
            assert.strictEqual(layer.graphics[0].infoTemplate.title, graphicData.infoTemplate.title, "Add GraphicFeature To Layer doesn't save expected infoTemplate title.");
            assert.strictEqual(layer.graphics[0].infoTemplate.content, graphicData.infoTemplate.content, "Add GraphicFeature To Layer doesn't save expected infoTemplate title.");

            // Add second Graphic Object
            polygonJson = {
                rings: [
                    [[-97.06138, 32.837], [-97.06133, 32.836], [-97.06124, 32.834], [-97.06127, 32.832], [-97.06138, 32.837]],
                    [[-97.06326, 32.759], [-97.06298, 32.755], [-97.06153, 32.749], [-97.06326, 32.759]]
                ],
                spatialReference: { wkid: 1110 }
            };
            polygon = new Polygon(polygonJson);
            graphicData = {
                geometry: polygon,
                attributes: { id: null, stringField: "Pond Test", doubleField: 508.54 },
                symbol: { color: [0, 110, 45, 255], width: 3, type: "esriSLS", style: "esriSLSSolid" },
                infoTemplate: { title: "Second Title", content: "Second Content" }
            };
            graphic = new Graphic(graphicData);
            idHelper = 'BBD8235F-886E-A8C2-8F2A-CDF17F0AABB9';
            symbol = new SimpleFillSymbol(SimpleFillSymbol.STYLE_SOLID,
                new SimpleLineSymbol(
                    SimpleLineSymbol.STYLE_SOLID,
                    new Color([123, 222, 112, 0.75]),
                    3),
                new Color([123, 210, 211, 0.5]));

            mapUtils.addGraphicFeatureToLayer(layer, graphic, symbol, idHelper);

            assert.strictEqual(layer.graphics.length, 2, "Add GraphicFeature To Layer doesn't save second graphic object.");
            assert.isTrue(layer.graphics[1] instanceof Graphic, "Add GraphicFeature To Layer doesn't save expected object.");
            assert.deepEqual(layer.graphics[1].geometry.rings, polygon.rings, "Add GraphicFeature To Layer doesn't save expected geometry paths.");
            assert.deepEqual(layer.graphics[1].geometry.spatialReference.wkid, polygon.spatialReference.wkid,
                "Add GraphicFeature To Layer doesn't save expected geometry spatialReference.");
            assert.deepEqual(layer.graphics[1].symbol, symbol,
                "Add GraphicFeature To Layer doesn't save expected symbol.");
            assert.strictEqual(layer.graphics[1].attributes.id, idHelper, "Add GraphicFeature To Layer doesn't save expected idHelper.");
            assert.strictEqual(layer.graphics[1].attributes.stringField, graphicData.attributes.stringField, "Add GraphicFeature To Layer doesn't save expected POND_NAME.");
            assert.strictEqual(layer.graphics[1].attributes.doubleField, graphicData.attributes.doubleField, "Add GraphicFeature To Layer doesn't save expected HECTARES.");
            assert.strictEqual(layer.graphics[1].infoTemplate.title, graphicData.infoTemplate.title, "Add GraphicFeature To Layer doesn't save expected infoTemplate title.");
            assert.strictEqual(layer.graphics[1].infoTemplate.content, graphicData.infoTemplate.content, "Add GraphicFeature To Layer doesn't save expected infoTemplate title.");
        }
    });
});