﻿define([
    'intern!object',
    'intern/chai!assert',
    'app/utils/Bookmarker',
    'app/utils/SimpleEncoder'
], function (registerSuite, assert, Bookmarker, encoder) {
    'use strict';
    registerSuite({
        name: 'Bookmarker module tests',
        setup: function () {
            //preserve original query string before any tests run so we can restore it in the teardown
            this.originalQueryString = location.href.substring(location.href.indexOf("?"), location.href.length);
        },
        //beforeEach: function () {},
        afterEach: function () {
            //this removes the existing event handler added during the previous test
            History.Adapter.handlers[History.Adapter.uid(window)].statechange = null;
        },
        teardown: function () {
            //need to restore the original query string
            //use the HTML5 history api to replace the current state (this doesn't create a new history and add it to the stack)
            window.history.pushState(null, null, window.location.pathname + this.originalQueryString);
        },

        'Test Bookmarker constructor with all options': function () {
            var parameter = "testState",
                bookmarker = new Bookmarker({
                    encode: false,
                    queryStringParameter: parameter,
                    getState: function() { return this.queryStringParameter + " getState"; },
                    onStateChange: function () { return this.queryStringParameter + " onStateChange"; },
                    restoreState: function () { return this.queryStringParameter + " restoreState"; }
                });
            assert.isFalse(bookmarker.encode, "encode should be false");
            assert.strictEqual(bookmarker.queryStringParameter, parameter, "query String Parameter name does not match expected value");
            assert.strictEqual(bookmarker.getState(), parameter + " getState", "query String Parameter name does not match expected value");
            assert.strictEqual(bookmarker.onStateChange(), parameter + " onStateChange", "query String Parameter name does not match expected value");
            assert.strictEqual(bookmarker.restoreState(), parameter + " restoreState", "query String Parameter name does not match expected value");
        },
        'Test Bookmarker constructor with default options': function () {
            var parameter = "state",
                bookmarker = new Bookmarker();
            assert.isTrue(bookmarker.encode, "encode should be true by default");
            assert.strictEqual(bookmarker.queryStringParameter, parameter, "query String Parameter name does not match expected value");
            assert.isTrue(bookmarker.cleanLoad, "cleanLoad should be true by default");
            assert.isFalse(bookmarker.justSavedState, "justSavedState should be false by default");
            assert.isFalse(bookmarker.justRestoredState, "justRestoredState should be false by default");
            assert.strictEqual(bookmarker.queryStringStub, "?state=", "default querySTringStub is not correct");
        },
        'Test Bookmarker default getState Function': function () {
            var bookmarker = new Bookmarker(),
                expectedGetState = {
                    message: "getState has not been defined by the application."
                };
            assert.deepEqual(bookmarker.getState(), expectedGetState, "incorrect message provided");
        },
        'Test Bookmarker default onStateChange Function': function () {
            var bookmarker = new Bookmarker();
            bookmarker.onStateChange();
            assert.isFalse(bookmarker.cleanLoad, "cleanLoad should be false after calling onStateChange");
            assert.isFalse(bookmarker.justSavedState, "justSavedState should be false by default");
        },
        'Test Bookmarker default restoreState Function': function () {
            var bookmarker = new Bookmarker(),
                testState = { fun: "woot!" },
                expectedRestoreState = {
                    message: "restoreState has not been defined by the application.",
                    state: testState
                },
                restoreResult, logEntry,
                originalLogFunction = window.console.log;

            window.console.log = function (object) { logEntry = object; };
            restoreResult = bookmarker.restoreState(testState);
            window.console.log = originalLogFunction; //restore logger
            assert.isFalse(restoreResult, "default restoreState function should return false");
            assert.deepEqual(logEntry, expectedRestoreState, "incorrect state returned");
        },
        'Test getStateFromQueryString with sample state': function () {
            var bookmarker = new Bookmarker({ encode: false }),
                queryStringState = null,
                stateObj = {
                extent: {
                    xmin: -14349441.539754221,
                    ymin: 6382616.894008949,
                    xmax: -10925062.672579234,
                    ymax: 8520407.70108819,
                    spatialReference: { wkid: 102100 }
                },
                basemap: "12",
                visibleLayers: ["layer0", "OS_Projects"]
                },
                encodedStateObj = encodeURIComponent(JSON.stringify(stateObj));
            
            //use the HTML5 history api to replace the current state (this doesn't create a new history and add it to the stack)
            window.history.pushState(stateObj, null, window.location.pathname + window.location.search + "&state=" + encodedStateObj);
            queryStringState = bookmarker.getStateFromQueryString();
            assert.deepEqual(queryStringState, stateObj, 'Retrieved State should equal beginning state.');
            //"pop" the state by going back. This will restore the querystring to 
            window.history.back();
        },
        'Test getStateFromQueryString with encoding': function () {
            var bookmarker = new Bookmarker({ encode: true }),
                queryStringState = null,
                stateObj = {
                    extent: {
                        xmin: -14349441.539754221,
                        ymin: 6382616.894008949,
                        xmax: -10925062.672579234,
                        ymax: 8520407.70108819,
                        spatialReference: { wkid: 102100 }
                    },
                    basemap: "12",
                    visibleLayers: ["layer0", "OS_Projects"]
                },
                encodedStateObj = encoder.encodeString(JSON.stringify(stateObj));

            //use the HTML5 history api to replace the current state (this doesn't create a new history and add it to the stack)
            window.history.pushState(stateObj, null, window.location.pathname + window.location.search + "&state=" + encodedStateObj);
            queryStringState = bookmarker.getStateFromQueryString();
            assert.deepEqual(queryStringState, stateObj, 'Retrieved State should equal beginning state.');
            //"pop" the state by going back. This will restore the querystring to 
            window.history.back();
        },
        'Test getStateFromQueryString with custom parameter': function () {
            var bookmarker = new Bookmarker({ encode: false }),
                queryStringState = null;
            bookmarker.queryStringParameter = "config";
            queryStringState = bookmarker.getStateFromQueryString();
            assert.deepEqual(queryStringState, "tests/intern", 'Retrieved object not what was expected.');
        },
        'Test saveState with cleanLoad': function() {
            var testObject = {
                    test1: "value1",
                    test2: "value2"
                },
                bookmarker = new Bookmarker({
                    encode: false,
                    getState: function () {
                        return testObject;
                    },
                    onStateChange: function() {
                        assert.fail("onStateChange should not be called during a cleanLoad");
                    }
                }),
                state;
            assert.isTrue(bookmarker.cleanLoad, "This test expects cleanLoad to be true");
            bookmarker.saveState();
            state = bookmarker.getStateFromHistory();
            assert.deepEqual(state.data, {}, "saveState should not return anything during a clean load");
            assert.isFalse(bookmarker.justRestoredState, "justRestoredState should be false after calling saveState");
        },
        'Test saveState after page load': function () {
            var testObject = {
                    test1: "value1",
                    test2: "value2"
                },
                bookmarker = new Bookmarker({
                    encode: false,
                    getState: function () {
                        return testObject;
                    },
                    onStateChange: function () {
                        var state = History.getState();
                        assert.deepEqual(state.data, testObject, "state should equal testObject");
                    }
                });
            bookmarker.cleanLoad = false;//saveState won't work while cleanLoad is true
            bookmarker.saveState();
            assert.isFalse(bookmarker.justRestoredState, "justRestoredState should be false after calling saveState");
            window.history.back();
        }

    });
});